//////////////////////////////////////////////////////////////////////////////
//
//  Copyright (C) 2007-2011 Infinite Z, Inc.  All Rights Reserved.
//
//  File:       ZSpaceInterface.cpp
//  Content:    Sample code demonstrating use of 
//				      ZSStereo's Head Tracking in Default (simple) mode.
//
//////////////////////////////////////////////////////////////////////////////

#include "ZSpaceInterface.h"
#include <freeglut.h>//"GL/glut.h"	//S3D

using namespace zspace::stereo;
using namespace zspace::tracker;

bool ZSpaceInterface::s_eyesInitialized = false;

bool
ZSpaceInterface::InitializeEyes(
#ifdef _WIN32
    HWND hWnd
#endif
    )
{
  if(!s_eyesInitialized)
  {
    // Initialize the stereoGL (for left-right frame synchronization)
    if (0 != zspace::stereo::StereoGl::initialize(hWnd))
    {
      // Failed to synchronize the left and right frame.
      ZS_LOG(ERR, "Failed to synchronize the left and right frame.\n" );
      return false;
    }
    s_eyesInitialized = true;
  }
  return true;
}


ZSpaceInterface::ZSpaceInterface()
: m_stereoWindow(NULL),
  m_stereoViewport(NULL),
  m_eyeMatricesPushed(false),
  m_isHeadTrackingEnabled(true)
{
  m_trackerSystem = new zspace::tracker::TrackerSystem();
}

ZSpaceInterface::~ZSpaceInterface()
{
  m_trackerSystem->removeReference();
}

bool ZSpaceInterface::initialize(int windowX, int windowY, int windowWidth, int windowHeight)
{
  // Initialize the stereo window and viewport.
  initializeStereo(windowX, windowY, windowWidth, windowHeight);

  return true;
}

void ZSpaceInterface::initializeStereo(int windowX, int windowY, int windowWidth, int windowHeight)
{
  // Create the stereo window.
  m_stereoWindow = StereoGl::createStereoWindow(windowX, windowY, windowWidth, windowHeight);

  // Create the stereo viewport.
  m_stereoViewport = StereoGl::createStereoViewport();
  m_stereoViewport->setUsingWindowSize(true);

   // Get the stereo frustum.
  StereoFrustum* stereoFrustum = m_stereoViewport->getStereoFrustum();
//  stereoFrustum->setStereoLevel(0.5f);						//S3D debug test zz
//stereoFrustum->setZeroParallaxOffset(-0.1f);
  stereoFrustum->setInterPupillaryDistance(0.065f);

//  printf ("StereoLevel: %f\n",stereoFrustum->getStereoLevel());			//S3D
//  printf ("zeroParallax: %f\n",stereoFrustum->getZeroParallaxOffset());
  printf ("IPD: %f\n",stereoFrustum->getInterPupillaryDistance());

  // Add the viewport to the stereo window.
  m_stereoWindow->addStereoViewport(m_stereoViewport);

}

void ZSpaceInterface::clean()
{
  if(m_stereoWindow)
  {
    m_stereoWindow->removeReference();
    m_stereoWindow = NULL;
  }
  if(m_stereoViewport)
  {
    m_stereoViewport->removeReference();
    m_stereoViewport = NULL;
  }
}

bool
ZSpaceInterface::getDefaultCameraPosition(float* pos3look3up3)
{
  const zspace::common::DisplayInfo::Display* display = m_stereoWindow->getCurrentDisplay();
  if(display)
  {
    pos3look3up3[0] = -display->displayOffset[0]; 
    pos3look3up3[1] = -display->displayOffset[1];
    pos3look3up3[2] = -display->displayOffset[2];
    pos3look3up3[3] = pos3look3up3[0] + display->displayOffset[0];
    pos3look3up3[4] = pos3look3up3[1] + display->displayOffset[1];
    pos3look3up3[5] = pos3look3up3[2] + display->displayOffset[2];
    pos3look3up3[6] = 0.0f;
    pos3look3up3[7] = 1.0f;
    pos3look3up3[8] = 0.0f;
  }
  return display != NULL;
}


void
ZSpaceInterface::setupStereoEye(StereoFrustum::StereoEye eye)
{
  if(m_eyeMatricesPushed)
    ZS_LOG(ERR, "Mismatched stereo eye setup: existing eye already configured\n" ); 
  pushStereoMatrices(eye);
  selectRenderContext(eye);
}

void
ZSpaceInterface::clearStereoEye()
{
  if(!m_eyeMatricesPushed)
    ZS_LOG(ERR, "Mismatched stereo eye setup: trying to clear nonexistent eye configuration\n" ); 
  popStereoMatrices();
  // don't have to set a new render context?
}

void
ZSpaceInterface::endStereo()
{
  m_stereoWindow->getStereoBuffer()->endStereo();
}


void
ZSpaceInterface::pushStereoMatrices(StereoFrustum::StereoEye eye)
{
  GLfloat monoModelViewGl[16];
  GLfloat viewMatrixGl[16];
  GLfloat projectionMatrixGl[16];

//quick hack: uncomment to render mono if you'd like
//eyeZSpace = StereoFrustum::STEREO_EYE_CENTER;
  int matrixMode;
  glGetIntegerv(GL_MATRIX_MODE, &matrixMode);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  // Copy the mono model-view matrix.
  glGetFloatv(GL_MODELVIEW_MATRIX, monoModelViewGl);

  // Get the stereo frustum.
  StereoFrustum* stereoFrustum = m_stereoViewport->getStereoFrustum();

  if (stereoFrustum)
  {
    // Modify the model-view matrix for eye.
    // This modification needs to be at the *bottom* of the matrix stack.
    // Note: The view matrix contains eye offset plus any rotation required for off-axis projection.
    stereoFrustum->getViewMatrix(eye, viewMatrixGl);

    glLoadMatrixf(viewMatrixGl);
    glMultMatrixf(monoModelViewGl);

    // Set the projection matrix for eye.
    // This matrix completely overwrites the existing projection matrix
    stereoFrustum->getProjectionMatrix(eye, projectionMatrixGl);

    glMatrixMode(GL_PROJECTION);
    glLoadMatrixf(projectionMatrixGl);
  }
  glMatrixMode(matrixMode);
}

void
ZSpaceInterface::popStereoMatrices()
{
  int matrixMode;
  glGetIntegerv(GL_MATRIX_MODE, &matrixMode);
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glMatrixMode(matrixMode);
}

void
ZSpaceInterface::selectRenderContext(StereoFrustum::StereoEye eye)
{
  if (eye == StereoFrustum::STEREO_EYE_LEFT)
    m_stereoWindow->getStereoBuffer()->selectRenderTarget(StereoBuffer::GRAPHICS_BUFFER_BACK_LEFT);
  else if(eye == StereoFrustum::STEREO_EYE_RIGHT)
    m_stereoWindow->getStereoBuffer()->selectRenderTarget(StereoBuffer::GRAPHICS_BUFFER_BACK_RIGHT);
}


void
ZSpaceInterface::captureTargets()
{
  // Get head tracker pose and pass it to the stereo frustum.
  m_trackerSystem->captureTargets();


  // Get head pose.
  zspace::common::Matrix4 headPose = zspace::common::Matrix4::IDENTITY();
  if (m_isHeadTrackingEnabled)
  {
    headPose = getHeadPose();
  }
  m_stereoViewport->getStereoFrustum()->setHeadPose(headPose);
}

void ZSpaceInterface::toggleHeadTracking()
{
  m_isHeadTrackingEnabled = !m_isHeadTrackingEnabled;
}

zspace::common::Matrix4
ZSpaceInterface::getHeadPose()
{
  // Get head pose.
  TrackerTarget* headTarget = m_trackerSystem->getDefaultHeadTarget();
  // NOTE: Function setHeadTrackingEnabled(false) of stereoFrustum actually discontinues updating head position. 
  // Setting an identity matrix actually bring the view into default head view. 
  if (headTarget && headTarget->isEnabled())
  {
    zspace::common::Matrix4 headPose;
    headTarget->getPose(headPose);
    return headPose;
  }
  return zspace::common::Matrix4::IDENTITY();
}

void
ZSpaceInterface::onMove(int x, int y)
{
  if (m_stereoWindow)
    m_stereoWindow->move(x, y);
}

void
ZSpaceInterface::onResize(int width, int height)
{
  if (m_stereoWindow)
  {
    m_stereoWindow->resize(width, height);
  }
}

void
ZSpaceInterface::onSwapRenderTarget()
{
  if (m_stereoWindow && m_stereoWindow->getStereoBuffer())
    m_stereoWindow->getStereoBuffer()->swapRenderTarget();
}

int
ZSpaceInterface::getWindowX() const
{
  if (m_stereoWindow)
    return m_stereoWindow->getX();
  return 0;
}

int
ZSpaceInterface::getWindowY() const
{
  if (m_stereoWindow)
    return m_stereoWindow->getY();
  return 0;
}

int
ZSpaceInterface::getWindowWidth() const
{
  if (m_stereoWindow)
    return m_stereoWindow->getWidth();
  return 0;
}

int
ZSpaceInterface::getWindowHeight() const
{
  if (m_stereoWindow)
    return m_stereoWindow->getHeight();
  return 0;
}
